@@ -26,6 +26,7 @@
 #include "index.h"
 #include "objects.h"
 
+class InputPlugin;
 class Plugin;
 class PluginHandle;
 class VFSFile;
@@ -37,7 +38,8 @@ typedef bool (* DirForeachFunc) (const char * path, const char * basename, void
 void adder_cleanup ();
 
 /* art.cc */
-void art_init ();
+void art_cache_current (const String & filename, Index<char> && data, String && art_file);
+void art_clear_current ();
 void art_cleanup ();
 
 /* art-search.cc */
@@ -52,6 +54,10 @@ void config_load ();
 void config_save ();
 void config_cleanup ();
 
+/* drct.cc */
+void record_init ();
+void record_cleanup ();
+
 /* effect.cc */
 void effect_start (int & channels, int & rate);
 Index<float> & effect_process (Index<float> & data);
@@ -90,11 +96,21 @@ void playback_play (int seek_time, bool pause);
 void playback_stop (bool exiting = false);
 
 bool playback_check_serial (int serial);
-bool playback_set_info (int entry, const String & filename,
- PluginHandle * decoder, Tuple && tuple);
+void playback_set_info (int entry, Tuple && tuple);
+void playback_setup_decode (const String & filename, InputPlugin * ip,
+ VFSFile && file, String && error);
 
-/* probe-buffer.cc */
-VFSFile probe_buffer_new (const char * filename);
+/* probe.cc */
+bool open_input_file (const char * filename, const char * mode,
+ InputPlugin * ip, VFSFile & file, String * error = nullptr);
+InputPlugin * load_input_plugin (PluginHandle * decoder, String * error = nullptr);
+
+/* internal versions of aud_file_* functions;
+ * these allow reuse of the same file handle during probing */
+PluginHandle * file_find_decoder (const char * filename, bool fast,
+ VFSFile & file, String * error = nullptr);
+bool file_read_tag (const char * filename, PluginHandle * decoder,
+ VFSFile & file, Tuple * tuple, Index<char> * image, String * error = nullptr);
 
 /* runtime.cc */
 extern size_t misc_bytes_allocated;
@@ -102,6 +118,9 @@ extern size_t misc_bytes_allocated;
 /* strpool.cc */
 void string_leak_check ();
 
+/* timer.cc */
+void timer_cleanup ();
+
 /* util.cc */
 const char * get_home_utf8 ();
 bool dir_foreach (const char * path, DirForeachFunc func, void * user_data);
@@ -112,6 +131,7 @@ const char * last_path_element (const char * path);
 void cut_path_element (char * path, int pos);
 
 unsigned int32_hash (unsigned val);
+unsigned ptr_hash (const void * ptr);
 
 /* vis-runner.cc */
 void vis_runner_start_stop (bool playing, bool paused);
