@@ -68,13 +68,15 @@ struct PlaybackControl {
 
 struct PlaybackInfo {
     // set by playback_set_info
-    String filename;
-    PluginHandle * decoder = nullptr;
-    Tuple tuple;
-
     int entry = -1;
+    Tuple tuple;
     String title;
 
+    // set by playback_setup_decode
+    String filename;
+    InputPlugin * ip = nullptr;
+    VFSFile file;
+
     // set by playback thread
     int length = -1;
     int time_offset = 0;
@@ -139,17 +141,12 @@ bool playback_check_serial (int serial)
 }
 
 // called from the playlist to update the tuple for the current song
-bool playback_set_info (int entry, const String & filename, PluginHandle * decoder, Tuple && tuple)
+void playback_set_info (int entry, Tuple && tuple)
 {
     // do nothing if the playback thread is lagging behind;
     // in that case, playback_set_info() will get called again anyway
     if (! lock_if (in_sync))
-        return false;
-
-    if (! pb_info.filename)
-        pb_info.filename = filename;
-    if (! pb_info.decoder)
-        pb_info.decoder = decoder;
+        return;
 
     if (tuple && tuple != pb_info.tuple)
     {
@@ -157,7 +154,10 @@ bool playback_set_info (int entry, const String & filename, PluginHandle * decod
 
         // don't call "tuple change" before "playback ready"
         if (is_ready ())
+        {
             event_queue ("tuple change", nullptr);
+            output_set_tuple (pb_info.tuple);
+        }
     }
 
     String title = pb_info.tuple.get_str (Tuple::FormattedTitle);
@@ -172,7 +172,21 @@ bool playback_set_info (int entry, const String & filename, PluginHandle * decod
     }
 
     unlock ();
-    return true;
+}
+
+// called from the playlist to set up decoding data
+void playback_setup_decode (const String & filename, InputPlugin * ip,
+ VFSFile && file, String && error)
+{
+    if (! lock_if (in_sync))
+        return;
+
+    pb_info.filename = filename;
+    pb_info.ip = ip;
+    pb_info.file = std::move (file);
+    pb_info.error_s = std::move (error);
+
+    unlock ();
 }
 
 // cleanup common to both playback_play() and playback_stop()
@@ -295,17 +309,20 @@ static void run_playback ()
 {
     // due to mutex ordering, we cannot call into the playlist while locked;
     // instead, playback_entry_read() calls back into playback_set_info()
-    if (! playback_entry_read (pb_state.playback_serial, pb_info.error_s))
+    playback_entry_read (pb_state.playback_serial);
+
+    if (! lock_if (in_sync))
+        return;
+
+    // check that we have all the necessary data
+    if (! pb_info.filename || ! pb_info.tuple || ! pb_info.ip ||
+     (! pb_info.ip->input_info.keys[InputKey::Scheme] && ! pb_info.file))
     {
         pb_info.error = true;
+        unlock ();
         return;
     }
 
-    lock ();
-
-    // playback_set_info() should always set this info
-    assert (pb_info.filename && pb_info.decoder && pb_info.tuple);
-
     // get various other bits of info from the tuple
     pb_info.length = pb_info.tuple.get_int (Tuple::Length);
     pb_info.time_offset = aud::max (0, pb_info.tuple.get_int (Tuple::StartTime));
@@ -318,29 +335,10 @@ static void run_playback ()
 
     unlock ();
 
-    InputPlugin * ip;
-    VFSFile file;
-
-    // load input plugin
-    if (! (ip = (InputPlugin *) aud_plugin_get_header (pb_info.decoder)))
-    {
-        pb_info.error = true;
-        pb_info.error_s = String (_("Error loading plugin"));
-        return;
-    }
-
-    // open file (not necessary for custom URI schemes)
-    if (! ip->input_info.keys[InputKey::Scheme] && ! (file = VFSFile (pb_info.filename, "r")))
-    {
-        pb_info.error = true;
-        pb_info.error_s = String (file.error ());
-        return;
-    }
-
     while (1)
     {
         // hand off control to input plugin
-        if (! ip->play (pb_info.filename, file))
+        if (! pb_info.ip->play (pb_info.filename, pb_info.file))
             pb_info.error = true;
 
         // close audio (no-op if it wasn't opened)
@@ -365,7 +363,7 @@ static void run_playback ()
             break;
 
         // rewind file pointer before repeating
-        if (file && file.fseek (0, VFS_SEEK_SET) != 0)
+        if (! open_input_file (pb_info.filename, "r", pb_info.ip, pb_info.file, & pb_info.error_s))
         {
             pb_info.error = true;
             break;
@@ -595,7 +593,7 @@ EXPORT void InputPlugin::set_playback_tuple (Tuple && tuple)
 {
     // due to mutex ordering, we cannot call into the playlist while locked;
     // instead, playback_entry_set_tuple() calls back into first
-    // playback_check_serial() and then eventually playlist_set_info()
+    // playback_check_serial() and then eventually playback_set_info()
     playback_entry_set_tuple (pb_state.playback_serial, std::move (tuple));
 }
 
@@ -634,6 +632,25 @@ EXPORT int InputPlugin::check_seek ()
     return seek;
 }
 
+/* compatibility (non-virtual) implementation of InputPlugin::read_tag(). */
+EXPORT bool InputPlugin::default_read_tag (const char * filename,
+ VFSFile & file, Tuple * tuple, Index<char> * image)
+{
+    /* just call read_tuple() and read_image() */
+    if (tuple)
+    {
+        if (! (* tuple = read_tuple (filename, file)))
+            return false;
+        if (image && file && file.fseek (0, VFS_SEEK_SET) != 0)
+            return true; /* true: tuple was read */
+    }
+
+    if (image)
+        * image = read_image (filename, file);
+
+    return true;
+}
+
 // thread-safe
 EXPORT bool aud_drct_get_playing ()
 {
