@@ -23,6 +23,7 @@
 #include "playlist-internal.h"
 #include "runtime.h"
 
+#include <assert.h>
 #include <pthread.h>
 #include <stdlib.h>
 #include <string.h>
@@ -65,6 +66,11 @@ enum PlaybackChange {
     PlaybackStopped
 };
 
+enum {
+    QueueChanged = (1 << 0),
+    DelayedUpdate = (1 << 1)
+};
+
 #define STATE_FILE "playlist-state"
 
 #define ENTER pthread_mutex_lock (& mutex)
@@ -104,8 +110,9 @@ struct Entry {
     Entry (PlaylistAddItem && item);
     ~Entry ();
 
+    void format ();
     void set_tuple (Tuple && new_tuple);
-    void set_failed (String && new_error);
+    void set_failed (const String & new_error);
 
     String filename;
     PluginHandle * decoder;
@@ -138,6 +145,13 @@ struct PlaylistData {
     int resume_time;
 };
 
+/* playback data provided by the playlist */
+struct PlaybackData {
+    InputPlugin * ip = nullptr;
+    VFSFile file;
+    String error;
+};
+
 static const char * const default_title = N_("New Playlist");
 static const char * const temp_title = N_("Now Playing");
 
@@ -159,22 +173,30 @@ static PlaylistData * playing_playlist = nullptr;
 static int resume_playlist = -1;
 static bool resume_paused = false;
 
+static bool metadata_fallbacks = false;
+static TupleCompiler title_formatter;
+
+static PlaybackData playback_data;
+
 static QueuedFunc queued_update;
 static UpdateLevel update_level;
+static bool update_delayed;
 
 struct ScanItem : public ListNode
 {
-    ScanItem (PlaylistData * playlist, Entry * entry, ScanRequest * request) :
+    ScanItem (PlaylistData * playlist, Entry * entry, ScanRequest * request, bool for_playback) :
         playlist (playlist),
         entry (entry),
-        request (request) {}
+        request (request),
+        for_playback (for_playback) {}
 
     PlaylistData * playlist;
     Entry * entry;
     ScanRequest * request;
+    bool for_playback;
 };
 
-static bool scan_enabled;
+static bool scan_enabled_nominal, scan_enabled;
 static int scan_playlist, scan_row;
 static List<ScanItem> scan_list;
 
@@ -188,7 +210,17 @@ static bool next_song_locked (PlaylistData * playlist, bool repeat, int hint);
 static void playlist_reformat_titles ();
 static void playlist_trigger_scan ();
 
-static SmartPtr<TupleCompiler> title_formatter;
+void Entry::format ()
+{
+    tuple.delete_fallbacks ();
+
+    if (metadata_fallbacks)
+        tuple.generate_fallbacks ();
+    else
+        tuple.generate_title ();
+
+    title_formatter.format (tuple);
+}
 
 void Entry::set_tuple (Tuple && new_tuple)
 {
@@ -204,17 +236,14 @@ void Entry::set_tuple (Tuple && new_tuple)
     if (! new_tuple)
         new_tuple.set_filename (filename);
 
-    new_tuple.generate_fallbacks ();
-    title_formatter->format (new_tuple);
-
     length = aud::max (0, new_tuple.get_int (Tuple::Length));
     tuple = std::move (new_tuple);
+
+    format ();
 }
 
 void PlaylistData::set_entry_tuple (Entry * entry, Tuple && tuple)
 {
-    scan_cancel (entry);
-
     total_length -= entry->length;
     if (entry->selected)
         selected_length -= entry->length;
@@ -226,11 +255,11 @@ void PlaylistData::set_entry_tuple (Entry * entry, Tuple && tuple)
         selected_length += entry->length;
 }
 
-void Entry::set_failed (String && new_error)
+void Entry::set_failed (const String & new_error)
 {
     scanned = true;
     failed = true;
-    error = std::move (new_error);
+    error = new_error;
 }
 
 Entry::Entry (PlaylistAddItem && item) :
@@ -322,22 +351,22 @@ static void update (void *)
 
     UpdateLevel level = update_level;
     update_level = NoUpdate;
+    update_delayed = false;
 
     LEAVE;
 
     hook_call ("playlist update", aud::to_ptr (level));
 }
 
-static bool send_playback_info (Entry * entry)
+static void send_playback_info (Entry * entry)
 {
     // if the entry was not scanned or failed to scan, we must still call
     // playback_set_info() in order to update the entry number
     Tuple tuple = (entry->scanned && ! entry->failed) ? entry->tuple.ref () : Tuple ();
-    return playback_set_info (entry->number, entry->filename, entry->decoder, std::move (tuple));
+    playback_set_info (entry->number, std::move (tuple));
 }
 
-static void queue_update (UpdateLevel level, PlaylistData * p, int at,
- int count, bool queue_changed = false)
+static void queue_update (UpdateLevel level, PlaylistData * p, int at, int count, int flags = 0)
 {
     if (p)
     {
@@ -365,15 +394,30 @@ static void queue_update (UpdateLevel level, PlaylistData * p, int at,
             p->next_update.after = p->entries.len () - at - count;
         }
 
-        if (queue_changed)
+        if ((flags & QueueChanged))
             p->next_update.queue_changed = true;
     }
 
     if (level == Structure)
         scan_restart ();
 
-    if (! update_level)
-        queued_update.queue (update, nullptr);
+    // only allow delayed update if a scan is still in progress
+    if ((flags & DelayedUpdate) && scan_enabled && p && (p->scanning || p->scan_ending))
+    {
+        if (! update_level)
+        {
+            queued_update.queue (250, update, nullptr);
+            update_delayed = true;
+        }
+    }
+    else
+    {
+        if (! update_level || update_delayed)
+        {
+            queued_update.queue (update, nullptr);
+            update_delayed = false;
+        }
+    }
 
     update_level = aud::max (update_level, level);
 }
@@ -457,13 +501,18 @@ static ScanItem * scan_list_find_request (ScanRequest * request)
     return nullptr;
 }
 
-static void scan_queue_entry (PlaylistData * playlist, Entry * entry)
+static void scan_queue_entry (PlaylistData * playlist, Entry * entry, bool for_playback = false)
 {
-    int flags = entry->scanned ? 0 : SCAN_TUPLE;
+    int flags = 0;
+    if (! entry->scanned)
+        flags |= SCAN_TUPLE;
+    if (for_playback)
+        flags |= (SCAN_IMAGE | SCAN_FILE);
+
     auto request = new ScanRequest (entry->filename, flags, scan_finish, entry->decoder);
     scanner_request (request);
 
-    scan_list.append (new ScanItem (playlist, entry, request));
+    scan_list.append (new ScanItem (playlist, entry, request, for_playback));
 }
 
 static void scan_check_complete (PlaylistData * playlist)
@@ -472,13 +521,20 @@ static void scan_check_complete (PlaylistData * playlist)
         return;
 
     playlist->scan_ending = false;
+
+    if (update_delayed)
+    {
+        queued_update.queue (update, nullptr);
+        update_delayed = false;
+    }
+
     event_queue_cancel ("playlist scan complete", nullptr);
     event_queue ("playlist scan complete", nullptr);
 }
 
 static bool scan_queue_next_entry ()
 {
-    if (! scan_enabled || aud_get_bool (nullptr, "metadata_on_play"))
+    if (! scan_enabled)
         return false;
 
     while (scan_playlist < playlists.len ())
@@ -491,7 +547,9 @@ static bool scan_queue_next_entry ()
             {
                 Entry * entry = playlist->entries[scan_row ++].get ();
 
-                if (! entry->scanned && ! scan_list_find_entry (entry))
+                // blacklist stdin
+                if (! entry->scanned && ! scan_list_find_entry (entry) &&
+                 strncmp (entry->filename, "stdin://", 8))
                 {
                     scan_queue_entry (playlist, entry);
                     return true;
@@ -539,7 +597,6 @@ static void scan_finish (ScanRequest * request)
     Entry * entry = item->entry;
 
     scan_list.remove (item);
-    delete item;
 
     if (! entry->decoder)
         entry->decoder = request->decoder;
@@ -547,11 +604,25 @@ static void scan_finish (ScanRequest * request)
     if (! entry->scanned && request->tuple)
     {
         playlist->set_entry_tuple (entry, std::move (request->tuple));
-        queue_update (Metadata, playlist, entry->number, 1);
+        queue_update (Metadata, playlist, entry->number, 1, DelayedUpdate);
     }
 
     if (! entry->decoder || ! entry->scanned)
-        entry->set_failed (std::move (request->error));
+        entry->set_failed (request->error);
+
+    if (item->for_playback && playlist == playing_playlist && entry == playlist->position)
+    {
+        // save playback data
+        playback_data.ip = request->ip;
+        playback_data.file = std::move (request->file);
+        playback_data.error = std::move (request->error);
+
+        // cache album art for current entry
+        art_cache_current (entry->filename, std::move (request->image_data),
+         std::move (request->image_file));
+    }
+
+    delete item;
 
     scan_check_complete (playlist);
     scan_schedule ();
@@ -614,7 +685,8 @@ static Entry * get_entry (int playlist_num, int entry_num,
         PlaylistData * playlist = lookup_playlist (playlist_num);
         Entry * entry = playlist ? lookup_entry (playlist, entry_num) : nullptr;
 
-        if (! entry || entry->failed)
+        // blacklist stdin
+        if (! entry || entry->failed || ! strncmp (entry->filename, "stdin://", 8))
             return entry;
 
         if ((need_decoder && ! entry->decoder) || (need_tuple && ! entry->scanned))
@@ -630,30 +702,19 @@ static Entry * get_entry (int playlist_num, int entry_num,
     }
 }
 
-/* mutex may be unlocked during the call */
-static Entry * get_playback_entry (int serial)
+static void start_playback (int seek_time, bool pause)
 {
-    while (1)
-    {
-        if (! playback_check_serial (serial))
-            return nullptr;
+    art_clear_current ();
+    playback_data = PlaybackData ();
+    playback_play (seek_time, pause);
+    scan_queue_entry (playing_playlist, playing_playlist->position, true);
+}
 
-        Entry * entry = playing_playlist ? playing_playlist->position : nullptr;
-
-        if (! entry || entry->failed)
-            return entry;
-
-        if (! entry->decoder || ! entry->scanned)
-        {
-            if (! scan_list_find_entry (entry))
-                scan_queue_entry (playing_playlist, entry);
-
-            pthread_cond_wait (& cond, & mutex);
-            continue;
-        }
-
-        return entry;
-    }
+static void stop_playback ()
+{
+    art_clear_current ();
+    playback_data = PlaybackData ();
+    playback_stop ();
 }
 
 void playlist_init ()
@@ -667,11 +728,10 @@ void playlist_init ()
     ENTER;
 
     update_level = NoUpdate;
+    update_delayed = false;
     scan_enabled = false;
     scan_playlist = scan_row = 0;
 
-    title_formatter.capture (new TupleCompiler);
-
     LEAVE;
 
     /* initialize title formatter */
@@ -679,43 +739,45 @@ void playlist_init ()
 
     hook_associate ("set metadata_on_play", (HookFunction) playlist_trigger_scan, nullptr);
     hook_associate ("set generic_title_format", (HookFunction) playlist_reformat_titles, nullptr);
-    hook_associate ("set show_numbers_in_pl", (HookFunction) playlist_reformat_titles, nullptr);
     hook_associate ("set leading_zero", (HookFunction) playlist_reformat_titles, nullptr);
+    hook_associate ("set metadata_fallbacks", (HookFunction) playlist_reformat_titles, nullptr);
+    hook_associate ("set show_numbers_in_pl", (HookFunction) playlist_reformat_titles, nullptr);
 }
 
 void playlist_enable_scan (bool enable)
 {
     ENTER;
 
-    if (! enable)
-        scan_list.clear ();
-
-    scan_enabled = enable;
+    scan_enabled_nominal = enable;
+    scan_enabled = scan_enabled_nominal && ! aud_get_bool (nullptr, "metadata_on_play");
+    scan_restart ();
 
     LEAVE;
-
-    if (enable)
-        playlist_trigger_scan ();
 }
 
 void playlist_end ()
 {
     hook_dissociate ("set metadata_on_play", (HookFunction) playlist_trigger_scan);
     hook_dissociate ("set generic_title_format", (HookFunction) playlist_reformat_titles);
-    hook_dissociate ("set show_numbers_in_pl", (HookFunction) playlist_reformat_titles);
     hook_dissociate ("set leading_zero", (HookFunction) playlist_reformat_titles);
+    hook_dissociate ("set metadata_fallbacks", (HookFunction) playlist_reformat_titles);
+    hook_dissociate ("set show_numbers_in_pl", (HookFunction) playlist_reformat_titles);
 
     ENTER;
 
+    /* playback should already be stopped */
+    assert (! playing_playlist);
+
     queued_update.stop ();
 
-    active_playlist = playing_playlist = nullptr;
+    active_playlist = nullptr;
     resume_playlist = -1;
+    resume_paused = false;
 
     playlists.clear ();
     unique_id_table.clear ();
 
-    title_formatter.clear ();
+    title_formatter.reset ();
 
     LEAVE;
 }
@@ -805,7 +867,7 @@ EXPORT void aud_playlist_delete (int playlist_num)
     if (playlist == playing_playlist)
     {
         playing_playlist = nullptr;
-        playback_stop ();
+        stop_playback ();
         was_playing = true;
     }
 
@@ -914,6 +976,14 @@ EXPORT int aud_playlist_get_active ()
     RETURN (list);
 }
 
+EXPORT int aud_playlist_new ()
+{
+    int playlist = aud_playlist_get_active () + 1;
+    aud_playlist_insert (playlist);
+    aud_playlist_set_active (playlist);
+    return playlist;
+}
+
 EXPORT void aud_playlist_play (int playlist_num, bool paused)
 {
     ENTER;
@@ -945,9 +1015,9 @@ EXPORT void aud_playlist_play (int playlist_num, bool paused)
     playing_playlist = playlist;
 
     if (playlist)
-        playback_play (playlist->resume_time, paused);
+        start_playback (playlist->resume_time, paused);
     else
-        playback_stop ();
+        stop_playback ();
 
     LEAVE;
 
@@ -975,10 +1045,7 @@ EXPORT int aud_playlist_get_blank ()
     String title = aud_playlist_get_title (list);
 
     if (strcmp (title, _(default_title)) || aud_playlist_entry_count (list) > 0)
-    {
-        list = aud_playlist_count ();
-        aud_playlist_insert (list);
-    }
+        aud_playlist_insert (++ list);
 
     return list;
 }
@@ -1017,13 +1084,13 @@ static PlaybackChange change_playback (PlaylistData * playlist)
 
     if (playlist->position)
     {
-        playback_play (0, aud_drct_get_paused ());
+        start_playback (0, aud_drct_get_paused ());
         return NextSong;
     }
     else
     {
         playing_playlist = nullptr;
-        playback_stop ();
+        stop_playback ();
         return PlaybackStopped;
     }
 }
@@ -1082,7 +1149,8 @@ EXPORT void aud_playlist_entry_delete (int playlist_num, int at, int number)
     ENTER_GET_PLAYLIST ();
 
     int entries = playlist->entries.len ();
-    bool position_changed = false, queue_changed = false;
+    bool position_changed = false;
+    int update_flags = 0;
     PlaybackChange change = NoChange;
 
     if (at < 0 || at > entries)
@@ -1115,7 +1183,7 @@ EXPORT void aud_playlist_entry_delete (int playlist_num, int at, int number)
         if (entry->queued)
         {
             playlist->queued.remove (playlist->queued.find (entry), 1);
-            queue_changed = true;
+            update_flags |= QueueChanged;
         }
 
         if (entry->selected)
@@ -1138,7 +1206,7 @@ EXPORT void aud_playlist_entry_delete (int playlist_num, int at, int number)
         change = change_playback (playlist);
     }
 
-    queue_update (Structure, playlist, at, 0, queue_changed);
+    queue_update (Structure, playlist, at, 0, update_flags);
     LEAVE;
 
     if (position_changed)
@@ -1216,6 +1284,10 @@ EXPORT void aud_playlist_set_focus (int playlist_num, int entry_num)
 {
     ENTER_GET_PLAYLIST ();
 
+    Entry * new_focus = lookup_entry (playlist, entry_num);
+    if (new_focus == playlist->focus)
+        RETURN ();
+
     int first = INT_MAX;
     int last = -1;
 
@@ -1225,7 +1297,7 @@ EXPORT void aud_playlist_set_focus (int playlist_num, int entry_num)
         last = aud::max (last, playlist->focus->number);
     }
 
-    playlist->focus = lookup_entry (playlist, entry_num);
+    playlist->focus = new_focus;
 
     if (playlist->focus)
     {
@@ -1420,7 +1492,8 @@ EXPORT void aud_playlist_delete_selected (int playlist_num)
         RETURN ();
 
     int entries = playlist->entries.len ();
-    bool position_changed = false, queue_changed = false;
+    bool position_changed = false;
+    int update_flags = 0;
     PlaybackChange change = NoChange;
 
     if (playlist->position && playlist->position->selected)
@@ -1448,7 +1521,7 @@ EXPORT void aud_playlist_delete_selected (int playlist_num)
             if (entry->queued)
             {
                 playlist->queued.remove (playlist->queued.find (entry), 1);
-                queue_changed = true;
+                update_flags |= QueueChanged;
             }
 
             playlist->total_length -= entry->length;
@@ -1476,7 +1549,7 @@ EXPORT void aud_playlist_delete_selected (int playlist_num)
         change = change_playback (playlist);
     }
 
-    queue_update (Structure, playlist, before, entries - after - before, queue_changed);
+    queue_update (Structure, playlist, before, entries - after - before, update_flags);
     LEAVE;
 
     if (position_changed)
@@ -1692,13 +1765,13 @@ static void playlist_reformat_titles ()
 {
     ENTER;
 
-    String format = aud_get_str (nullptr, "generic_title_format");
-    title_formatter->compile (format);
+    metadata_fallbacks = aud_get_bool (nullptr, "metadata_fallbacks");
+    title_formatter.compile (aud_get_str (nullptr, "generic_title_format"));
 
     for (auto & playlist : playlists)
     {
         for (auto & entry : playlist->entries)
-            title_formatter->format (entry->tuple);
+            entry->format ();
 
         queue_update (Metadata, playlist.get (), 0, playlist->entries.len ());
     }
@@ -1710,9 +1783,7 @@ static void playlist_trigger_scan ()
 {
     ENTER;
 
-    for (auto & playlist : playlists)
-        scan_queue_playlist (playlist.get ());
-
+    scan_enabled = scan_enabled_nominal && ! aud_get_bool (nullptr, "metadata_on_play");
     scan_restart ();
 
     LEAVE;
@@ -1731,6 +1802,7 @@ static void playlist_rescan_real (int playlist_num, bool selected)
     queue_update (Metadata, playlist, 0, playlist->entries.len ());
     scan_queue_playlist (playlist);
     scan_restart ();
+
     LEAVE;
 }
 
@@ -1815,7 +1887,7 @@ EXPORT void aud_playlist_queue_insert (int playlist_num, int at, int entry_num)
 
     entry->queued = true;
 
-    queue_update (Selection, playlist, entry_num, 1, true);
+    queue_update (Selection, playlist, entry_num, 1, QueueChanged);
     LEAVE;
 }
 
@@ -1844,7 +1916,7 @@ EXPORT void aud_playlist_queue_insert_selected (int playlist_num, int at)
     playlist->queued.move_from (add, 0, at, -1, true, true);
 
     if (first < playlist->entries.len ())
-        queue_update (Selection, playlist, first, last + 1 - first, true);
+        queue_update (Selection, playlist, first, last + 1 - first, QueueChanged);
 
     LEAVE;
 }
@@ -1888,7 +1960,7 @@ EXPORT void aud_playlist_queue_delete (int playlist_num, int at, int number)
     playlist->queued.remove (at, number);
 
     if (first < entries)
-        queue_update (Selection, playlist, first, last + 1 - first, true);
+        queue_update (Selection, playlist, first, last + 1 - first, QueueChanged);
 
     LEAVE;
 }
@@ -1916,7 +1988,7 @@ EXPORT void aud_playlist_queue_delete_selected (int playlist_num)
     }
 
     if (first < entries)
-        queue_update (Selection, playlist, first, last + 1 - first, true);
+        queue_update (Selection, playlist, first, last + 1 - first, QueueChanged);
 
     LEAVE;
 }
@@ -1968,45 +2040,89 @@ bool playlist_prev_song (int playlist_num)
 
 static bool shuffle_next (PlaylistData * playlist)
 {
-    int choice = 0;
-    Entry * found = nullptr;
+    bool by_album = aud_get_bool (nullptr, "album_shuffle");
 
-    for (auto & entry : playlist->entries)
+    // helper #1: determine whether two entries are in the same album
+    auto same_album = [] (const Tuple & a, const Tuple & b)
     {
-        if (! entry->shuffle_num)
-            choice ++;
-        else if (playlist->position &&
-         entry->shuffle_num > playlist->position->shuffle_num &&
-         (! found || entry->shuffle_num < found->shuffle_num))
-            found = entry.get ();
-    }
+        String album = a.get_str (Tuple::Album);
+        return (album && album == b.get_str (Tuple::Album));
+    };
 
-    if (found)
+    // helper #2: determine whether an entry is among the shuffle choices
+    auto is_choice = [&] (Entry * prev, Entry * entry)
     {
-        set_position (playlist, found, false);
-        return true;
-    }
+        return (! entry->shuffle_num) && (! by_album || ! prev ||
+         prev->shuffle_num || ! same_album (prev->tuple, entry->tuple));
+    };
 
-    if (! choice)
-        return false;
-
-    choice = rand () % choice;
-
-    for (auto & entry : playlist->entries)
+    if (playlist->position)
     {
-        if (! entry->shuffle_num)
+        // step #1: check to see if the shuffle order is already established
+        Entry * next = nullptr;
+
+        for (auto & entry : playlist->entries)
         {
-            if (! choice)
-            {
-                set_position (playlist, entry.get (), true);
-                break;
-            }
+            if (entry->shuffle_num > playlist->position->shuffle_num &&
+             (! next || entry->shuffle_num < next->shuffle_num))
+                next = entry.get ();
+        }
 
-            choice --;
+        if (next)
+        {
+            set_position (playlist, next, false);
+            return true;
+        }
+
+        // step #2: check to see if we should advance to the next entry
+        if (by_album && playlist->position->number + 1 < playlist->entries.len ())
+        {
+            next = playlist->entries[playlist->position->number + 1].get ();
+
+            if (! next->shuffle_num && same_album (playlist->position->tuple, next->tuple))
+            {
+                set_position (playlist, next, true);
+                return true;
+            }
         }
     }
 
-    return true;
+    // step #3: count the number of possible shuffle choices
+    int choices = 0;
+    Entry * prev = nullptr;
+
+    for (auto & entry : playlist->entries)
+    {
+        if (is_choice (prev, entry.get ()))
+            choices ++;
+
+        prev = entry.get ();
+    }
+
+    if (! choices)
+        return false;
+
+    // step #4: pick one of those choices by random and find it again
+    choices = rand () % choices;
+    prev = nullptr;
+
+    for (auto & entry : playlist->entries)
+    {
+        if (is_choice (prev, entry.get ()))
+        {
+            if (! choices)
+            {
+                set_position (playlist, entry.get (), true);
+                return true;
+            }
+
+            choices --;
+        }
+
+        prev = entry.get ();
+    }
+
+    return false;  // never reached
 }
 
 static void shuffle_reset (PlaylistData * playlist)
@@ -2029,7 +2145,7 @@ static bool next_song_locked (PlaylistData * playlist, bool repeat, int hint)
         playlist->queued.remove (0, 1);
         playlist->position->queued = false;
 
-        queue_update (Selection, playlist, playlist->position->number, 1, true);
+        queue_update (Selection, playlist, playlist->position->number, 1, QueueChanged);
     }
     else if (aud_get_bool (nullptr, "shuffle"))
     {
@@ -2078,32 +2194,46 @@ bool playlist_next_song (int playlist_num, bool repeat)
     return true;
 }
 
-bool playback_entry_read (int serial, String & error)
+static Entry * get_playback_entry (int serial)
 {
-    ENTER;
-    bool success = false;
-    Entry * entry = get_playback_entry (serial);
+    if (! playing_playlist || ! playback_check_serial (serial))
+        return nullptr;
 
-    if (entry && send_playback_info (entry))
-    {
-        success = ! entry->failed;
-        error = entry->error;
-    }
-
-    RETURN (success);
+    return playing_playlist->position;
 }
 
+// called from playback thread
+void playback_entry_read (int serial)
+{
+    ENTER;
+    Entry * entry;
+
+    // wait for the playback entry to be scanned (or deleted)
+    while ((entry = get_playback_entry (serial)) && scan_list_find_entry (entry))
+        pthread_cond_wait (& cond, & mutex);
+
+    if (entry)
+    {
+        send_playback_info (entry);
+        playback_setup_decode (entry->filename, playback_data.ip,
+         std::move (playback_data.file), std::move (playback_data.error));
+    }
+
+    LEAVE;
+}
+
+// called from playback thread
 void playback_entry_set_tuple (int serial, Tuple && tuple)
 {
     ENTER;
-    PlaylistData * playlist = playing_playlist;
-    if (! playlist || ! playlist->position || ! playback_check_serial (serial))
-        RETURN ();
+    Entry * entry = get_playback_entry (serial);
 
-    Entry * entry = playlist->position;
-    playlist->set_entry_tuple (entry, std::move (tuple));
+    if (entry)
+    {
+        playing_playlist->set_entry_tuple (entry, std::move (tuple));
+        queue_update (Metadata, playing_playlist, entry->number, 1);
+    }
 
-    queue_update (Metadata, playlist, entry->number, 1);
     LEAVE;
 }
 
@@ -2238,16 +2368,30 @@ void playlist_load_state ()
 
     fclose (handle);
 
+    /* set initial focus and selection */
     /* clear updates queued during init sequence */
 
     for (auto & playlist : playlists)
     {
+        Entry * focus = playlist->position;
+        if (! focus && playlist->entries.len ())
+            focus = playlist->entries[0].get ();
+
+        if (focus)
+        {
+            focus->selected = true;
+            playlist->focus = focus;
+            playlist->selected_count = 1;
+            playlist->selected_length = focus->length;
+        }
+
         playlist->next_update = Update ();
         playlist->last_update = Update ();
     }
 
     queued_update.stop ();
     update_level = NoUpdate;
+    update_delayed = false;
 
     LEAVE;
 }
