@@ -191,11 +191,16 @@ static void create_spin_button (const PreferencesWidget * widget,
  const char * domain)
 {
     * label_pre = gtk_label_new (dgettext (domain, widget->label));
+    gtk_misc_set_alignment ((GtkMisc *) * label_pre, 1, 0.5);
+
     * spin_btn = gtk_spin_button_new_with_range (widget->data.spin_btn.min,
      widget->data.spin_btn.max, widget->data.spin_btn.step);
 
     if (widget->data.spin_btn.right_label)
+    {
         * label_past = gtk_label_new (dgettext (domain, widget->data.spin_btn.right_label));
+        gtk_misc_set_alignment ((GtkMisc *) * label_past, 0, 0.5);
+    }
 
     widget_init (* spin_btn, widget);
 }
@@ -292,7 +297,10 @@ static void create_cbox (const PreferencesWidget * widget, GtkWidget * * label,
     * combobox = gtk_combo_box_text_new ();
 
     if (widget->label)
+    {
         * label = gtk_label_new (dgettext (domain, widget->label));
+        gtk_misc_set_alignment ((GtkMisc *) * label, 1, 0.5);
+    }
 
     g_object_set_data ((GObject *) * combobox, "combodomain", (void *) domain);
     widget_init (* combobox, widget);
@@ -321,7 +329,6 @@ static void fill_table (GtkWidget * table,
 
             case PreferencesWidget::FontButton:
                 create_font_btn (& w, & widget_left, & widget_middle, domain);
-                middle_policy = (GtkAttachOptions) (GTK_EXPAND | GTK_FILL);
                 break;
 
             case PreferencesWidget::Entry:
@@ -331,7 +338,6 @@ static void fill_table (GtkWidget * table,
 
             case PreferencesWidget::ComboBox:
                 create_cbox (& w, & widget_left, & widget_middle, domain);
-                middle_policy = (GtkAttachOptions) (GTK_EXPAND | GTK_FILL);
                 break;
 
             default:
@@ -361,8 +367,12 @@ void audgui_create_widgets_with_domain (GtkWidget * box,
  ArrayRef<PreferencesWidget> widgets, const char * domain)
 {
     GtkWidget * widget = nullptr, * child_box = nullptr;
+    bool disable_child = false;
     GSList * radio_btn_group = nullptr;
 
+    int indent = 0;
+    int spacing = 0;
+
     for (const PreferencesWidget & w : widgets)
     {
         GtkWidget * label = nullptr;
@@ -379,19 +389,19 @@ void audgui_create_widgets_with_domain (GtkWidget * box,
                 gtk_alignment_set_padding ((GtkAlignment *) alignment, 0, 0, 12, 0);
                 gtk_container_add ((GtkContainer *) alignment, child_box);
 
-                if (GTK_IS_TOGGLE_BUTTON (widget))
-                    gtk_widget_set_sensitive (child_box,
-                     gtk_toggle_button_get_active ((GtkToggleButton *) widget));
+                if (disable_child)
+                    gtk_widget_set_sensitive (child_box, false);
             }
         }
         else
             child_box = nullptr;
 
         GtkWidget * alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
-        gtk_alignment_set_padding ((GtkAlignment *) alignment, 6, 0, 12, 0);
+        gtk_alignment_set_padding ((GtkAlignment *) alignment, spacing, 0, indent, 0);
         gtk_box_pack_start ((GtkBox *) (child_box ? child_box : box), alignment, false, false, 0);
 
         widget = nullptr;
+        disable_child = false;
 
         if (radio_btn_group && w.type != PreferencesWidget::RadioButton)
             radio_btn_group = nullptr;
@@ -405,14 +415,20 @@ void audgui_create_widgets_with_domain (GtkWidget * box,
 
             case PreferencesWidget::CheckButton:
                 widget = gtk_check_button_new_with_mnemonic (dgettext (domain, w.label));
+                disable_child = ! w.cfg.get_bool ();
                 widget_init (widget, & w);
                 break;
 
             case PreferencesWidget::Label:
             {
                 if (strstr (w.label, "<b>"))
-                    gtk_alignment_set_padding ((GtkAlignment *) alignment,
-                     (& w == widgets.data) ? 0 : 12, 0, 0, 0);
+                {
+                    /* headings get double spacing and no indent */
+                    gtk_alignment_set_padding ((GtkAlignment *) alignment, 2 * spacing, 0, 0, 0);
+
+                    /* set indent for items below the heading */
+                    indent = 12;
+                }
 
                 GtkWidget * icon = nullptr;
                 create_label (& w, & label, & icon, domain);
@@ -558,13 +574,10 @@ void audgui_create_widgets_with_domain (GtkWidget * box,
         }
 
         if (widget)
-        {
-            /* use uniform spacing for horizontal boxes */
-            if (gtk_orientable_get_orientation ((GtkOrientable *) box) ==
-             GTK_ORIENTATION_HORIZONTAL)
-                gtk_alignment_set_padding ((GtkAlignment *) alignment, 0, 0, 0, 0);
-
             gtk_container_add ((GtkContainer *) alignment, widget);
-        }
+
+        /* wait till after first widget to set item spacing */
+        if (gtk_orientable_get_orientation ((GtkOrientable *) box) == GTK_ORIENTATION_VERTICAL)
+            spacing = 6;
     }
 }
